import sys

from ...exceptions import FastAgencyCLIPythonVersionError

if sys.version_info < (3, 10):
    raise FastAgencyCLIPythonVersionError("Mesop requires Python 3.10 or higher")

from .base import MesopUI

__all__ = ["MesopUI"]



from collections.abc import Iterable

import mesop as me

from ...base import Workflows
from .base import MesopMessage, MesopUI, run_workflow
from .data_model import State


def send_prompt_to_autogen(
    prompt: str, wf: Workflows, name: str
) -> Iterable[MesopMessage]:
    mesop_io = run_workflow(wf, name=name, initial_message=prompt)
    state = me.state(State)
    state.conversation.fastagency = mesop_io.id
    return mesop_io.get_message_stream()


def send_user_feedback_to_autogen(user_response: str) -> Iterable[MesopMessage]:
    state = me.state(State)
    mesop_id = state.conversation.fastagency
    mesop_io = MesopUI.get_conversation(mesop_id)
    mesop_io.respond(user_response)
    return mesop_io.get_message_stream()



from collections.abc import Iterator
from contextlib import contextmanager
from functools import cached_property
from typing import Optional, Union

from fastapi_code_generator.parser import (
    Argument,
    OpenAPIParser,
    ParameterObject,
    ReferenceObject,
)


class ArgumentWithDescription(Argument):  # type: ignore[misc]
    description: Optional[str] = None

    @cached_property
    def argument(self) -> str:
        if self.description:
            description = self.description.replace('"""', '"""')
            type_hint = f'Annotated[{self.type_hint}, """{description}"""]'
        else:
            type_hint = self.type_hint

        if self.default is None and self.required:
            return f"{self.name}: {type_hint}"

        return f"{self.name}: {type_hint} = {self.default}"


@contextmanager
def patch_get_parameter_type() -> Iterator[None]:
    original_get_parameter_type = OpenAPIParser.get_parameter_type

    def get_parameter_type(
        self: OpenAPIParser,
        parameters: Union[ReferenceObject, ParameterObject],
        snake_case: bool,
        path: list[str],
    ) -> Optional[Argument]:
        # get the original argument
        argument = original_get_parameter_type(self, parameters, snake_case, path)

        # add description to the argument
        parameters = self.resolve_object(parameters, ParameterObject)
        argument_with_description = ArgumentWithDescription(
            description=parameters.description, **argument.model_dump()
        )
        return argument_with_description

    OpenAPIParser.get_parameter_type = get_parameter_type

    try:
        yield
    finally:
        OpenAPIParser.get_parameter_type = original_get_parameter_type



from logging import getLogger
from pathlib import Path
from typing import Annotated, Optional

import typer

from .. import __version__
from ..exceptions import FastAgencyCLIError, FastAgencyCLIPythonVersionError
from .discover import get_import_string
from .logging import setup_logging

app = typer.Typer(rich_markup_mode="rich")

setup_logging()
logger = getLogger(__name__)


def version_callback(value: bool) -> None:
    if value:
        typer.echo(f"{__version__}")
        raise typer.Exit()


@app.callback()
def callback(
    version: Annotated[
        Optional[bool],
        typer.Option(
            "--version", help="Show the version and exit.", callback=version_callback
        ),
    ] = None,
) -> None:
    """FastAgency CLI - The [bold]fastapi[/bold] command line app. 😎

    Manage your [bold]FastAgency[/bold] projects, run your FastAgency apps, and more.

    Read more in the docs: [link]https://fastagency.ai/latest/[/link].
    """  # noqa: D415


def _run_app(
    *,
    path: Optional[Path],
    app: Optional[str],
    workflow: Optional[str],
    initial_message: Optional[str],
    dev_mode: bool = False,
    single_run: bool = False,
) -> None:
    try:
        import_string, fa_app = get_import_string(path=path, app_name=app)

        with fa_app.create(import_string=import_string):
            fa_app.start(
                import_string=import_string,
                name=workflow,
                initial_message=initial_message,
                single_run=single_run,
            )
    except FastAgencyCLIPythonVersionError as e:
        msg = e.args[0]
        typer.echo(msg, err=True)
        raise typer.Exit(code=1)  # noqa: B904
    except FastAgencyCLIError as e:
        msg = e.args[0]
        typer.echo(msg, err=True)
        raise typer.Exit(code=1) from None


def _get_help_messages(dev_mode: bool = False) -> dict[str, str]:
    help = f"""Run a [bold]FastAgency[/bold] app in [yellow]{'development' if dev_mode else 'production'}[/yellow] mode. 🚀

{'This is similar to the [bold]fastagency run[/bold] command but with [bold]reload[/bold] enabled and listening on the [blue]127.0.0.1[/blue] address.' if dev_mode else 'This is similar to the [bold]fastagency dev[/bold] command, but optimized for production environments.'}

It automatically detects the Python module or package that needs to be imported based on the file or directory path passed.

If no path is passed, it tries with:

- [blue]main.py[/blue]
- [blue]app.py[/blue]
- [blue]api.py[/blue]
- [blue]app/main.py[/blue]
- [blue]app/app.py[/blue]
- [blue]app/api.py[/blue]

It also detects the directory that needs to be added to the [bold]PYTHONPATH[/bold] to make the app importable and adds it.

It detects the [bold]FastAgency[/bold] app object to use. By default it looks in the module or package for an object named:

- [blue]app[/blue]
- [blue]api[/blue]

Otherwise, it uses the first [bold]FastAgency[/bold] app found in the imported module or package.
"""
    short_help = f"Run a [bold]FastAgency[/bold] app in [yellow]{'development' if dev_mode else 'production'}[/yellow] mode."
    return {"help": help, "short_help": short_help}


@app.command(**_get_help_messages(False))  # type: ignore[arg-type]
def run(
    path: Annotated[
        Optional[Path],
        typer.Argument(
            help="A path to a Python file or package directory (with [blue]__init__.py[/blue] files) containing a [bold]FastAgency[/bold] app. If not provided, a default set of paths will be tried."
        ),
    ] = None,
    *,
    app: Annotated[
        Optional[str],
        typer.Option(
            help="The name of the variable that contains the [bold][/bold] app in the imported module or package. If not provided, it is detected automatically."
        ),
    ] = None,
    workflow: Annotated[
        Optional[str],
        typer.Option(
            "--workflow",
            "-w",
            help="The name of the workflow to run. If not provided, the default workflow will be run.",
        ),
    ] = None,
    initial_message: Annotated[
        Optional[str],
        typer.Option(
            "--initial_message",
            "-i",
            help="The initial message to send to the workflow. If not provided, a default message will be sent.",
        ),
    ] = None,
    single_run: Annotated[
        bool,
        typer.Option(
            "--single-run", help="If set, only a single workflow will be executed."
        ),
    ] = False,
) -> None:
    dev_mode = False
    _run_app(
        path=path,
        app=app,
        workflow=workflow,
        initial_message=initial_message,
        dev_mode=dev_mode,
        single_run=single_run,
    )


@app.command(**_get_help_messages(True))  # type: ignore[arg-type]
def dev(
    path: Annotated[
        Optional[Path],
        typer.Argument(
            help="A path to a Python file or package directory (with [blue]__init__.py[/blue] files) containing a [bold]FastAgency[/bold] app. If not provided, a default set of paths will be tried."
        ),
    ] = None,
    *,
    app: Annotated[
        Optional[str],
        typer.Option(
            help="The name of the variable that contains the [bold][/bold] app in the imported module or package. If not provided, it is detected automatically."
        ),
    ] = None,
    workflow: Annotated[
        Optional[str],
        typer.Option(
            "--workflow",
            "-w",
            help="The name of the workflow to run. If not provided, the default workflow will be run.",
        ),
    ] = None,
    initial_message: Annotated[
        Optional[str],
        typer.Option(
            "--initial_message",
            "-i",
            help="The initial message to send to the workflow. If not provided, a default message will be sent.",
        ),
    ] = None,
    single_run: Annotated[
        bool,
        typer.Option(
            "--single-run", help="If set, only a single workflow will be executed."
        ),
    ] = False,
) -> None:
    dev_mode = True
    _run_app(
        path=path,
        app=app,
        workflow=workflow,
        initial_message=initial_message,
        dev_mode=dev_mode,
        single_run=single_run,
    )


@app.command(help="Display the version of FastAgency")
def version() -> None:
    typer.echo(__version__)


def main() -> None:
    app()



"""CLI entry point to FastAgency library."""

import warnings

from .cli import app as cli

warnings.filterwarnings("default", category=ImportWarning, module="fastagency")

if __name__ == "__main__":
    cli(prog_name="fastagency")



import os
from typing import Annotated, Any, Optional, Union

from autogen.agentchat import AssistantAgent as AutoGenAssistantAgent
from autogen.agentchat import ConversableAgent as AutoGenConversableAgent
from autogen.agentchat.chat import ChatResult
from autogen.agentchat.contrib.web_surfer import WebSurferAgent as AutoGenWebSurferAgent
from pydantic import BaseModel, Field, HttpUrl

__all__ = ["WebSurferAnswer", "WebSurferTool"]


class WebSurferAnswer(BaseModel):
    task: Annotated[str, Field(..., description="The task to be completed")]
    is_successful: Annotated[
        bool, Field(..., description="Whether the task was successful")
    ]
    short_answer: Annotated[
        str,
        Field(
            ...,
            description="The short answer to the task without any explanation",
        ),
    ]
    long_answer: Annotated[
        str,
        Field(..., description="The long answer to the task with explanation"),
    ]
    visited_links: Annotated[
        list[HttpUrl],
        Field(..., description="The list of visited links to generate the answer"),
    ]

    @staticmethod
    def get_example_answer() -> "WebSurferAnswer":
        return WebSurferAnswer(
            task="What is the most popular QLED TV to buy on amazon.com?",
            is_successful=True,
            short_answer='Amazon Fire TV 55" Omni QLED Series 4K UHD smart TV',
            long_answer='Amazon has the best selling page by different categories and there is a category for QLED TVs under electroincs. The most popular QLED TV is Amazon Fire TV 55" Omni QLED Series 4K UHD smart TV, Dolby Vision IQ, Fire TV Ambient Experience, local dimming, hands-free with Alexa. It is the best selling QLED TV on Amazon.',
            visited_links=[
                "https://www.amazon.com/Best-Sellers/",
                "https://www.amazon.com/Best-Sellers-Electronics-QLED-TVs/",
            ],
        )


class WebSurferTool:  # implements Toolable
    def __init__(
        self,
        *,
        name_prefix: str,
        llm_config: dict[str, Any],
        summarizer_llm_config: dict[str, Any],
        viewport_size: int = 4096,
        bing_api_key: Optional[str] = None,
        max_consecutive_auto_reply: int = 30,
        max_links_to_click: int = 10,
        websurfer_kwargs: Optional[dict[str, Any]] = None,
        assistant_kwargs: Optional[dict[str, Any]] = None,
    ):
        """Create a new WebSurferChat instance.

        Args:
            name_prefix (str): The name prefix of the inner AutoGen agents
            llm_config (Dict[str, Any]): The LLM configuration
            summarizer_llm_config (Dict[str, Any]): The summarizer LLM configuration
            viewport_size (int, optional): The viewport size. Defaults to 4096.
            bing_api_key (Optional[str], optional): The Bing API key. Defaults to None.
            max_consecutive_auto_reply (int, optional): The maximum consecutive auto reply. Defaults to 30.
            max_links_to_click (int, optional): The maximum links to click. Defaults to 10.
            websurfer_kwargs (Optional[Dict[str, Any]], optional): The WebSurfer kwargs. Defaults to None.
            assistant_kwargs (Optional[Dict[str, Any]], optional): The Assistant kwargs. Defaults to None.
        """
        if websurfer_kwargs is None:
            websurfer_kwargs = {}
        if assistant_kwargs is None:
            assistant_kwargs = {}

        self.name_prefix = name_prefix
        self.llm_config = llm_config
        self.summarizer_llm_config = summarizer_llm_config
        self.viewport_size = viewport_size
        self.bing_api_key = (
            bing_api_key if bing_api_key is not None else os.getenv("BING_API_KEY")
        )
        self.max_consecutive_auto_reply = max_consecutive_auto_reply
        self.max_links_to_click = max_links_to_click
        self.websurfer_kwargs = websurfer_kwargs
        self.assistant_kwargs = assistant_kwargs

        self.task = "not set yet"
        self.last_is_termination_msg_error = ""

        self.browser_config = {
            "viewport_size": self.viewport_size,
            "bing_api_key": self.bing_api_key,
            "request_kwargs": {
                "headers": {
                    "User-Agent": "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/41.0.2228.0 Safari/537.36",
                }
            },
        }

        if "human_input_mode" in self.websurfer_kwargs:
            self.websurfer_kwargs.pop("human_input_mode")

        self.websurfer = AutoGenWebSurferAgent(
            name=f"{self.name_prefix}_inner_websurfer",
            llm_config=self.llm_config,
            summarizer_llm_config=self.summarizer_llm_config,
            browser_config=self.browser_config,
            human_input_mode="NEVER",
            is_termination_msg=self.is_termination_msg,
            **self.websurfer_kwargs,
        )

        if "human_input_mode" in self.assistant_kwargs:
            self.assistant_kwargs.pop("human_input_mode")

        self.assistant = AutoGenAssistantAgent(
            name=f"{self.name_prefix}_inner_assistant",
            llm_config=self.llm_config,
            human_input_mode="NEVER",
            system_message=self.system_message,
            max_consecutive_auto_reply=self.max_consecutive_auto_reply,
            # is_termination_msg=self.is_termination_msg,
            **self.assistant_kwargs,
        )

    def is_termination_msg(self, msg: dict[str, Any]) -> bool:
        # print(f"is_termination_msg({msg=})")
        if (
            "content" in msg
            and msg["content"] is not None
            and "TERMINATE" in msg["content"]
        ):
            return True
        try:
            WebSurferAnswer.model_validate_json(msg["content"])
            return True
        except Exception as e:
            self.last_is_termination_msg_error = str(e)
            return False

    def _get_error_message(self, chat_result: ChatResult) -> Optional[str]:
        messages = [msg["content"] for msg in chat_result.chat_history]
        last_message = messages[-1]
        if "TERMINATE" in last_message:
            return self.error_message

        try:
            WebSurferAnswer.model_validate_json(last_message)
        except Exception:
            return self.error_message

        return None

    def _get_answer(self, chat_result: ChatResult) -> WebSurferAnswer:
        messages = [msg["content"] for msg in chat_result.chat_history]
        last_message = messages[-1]
        return WebSurferAnswer.model_validate_json(last_message)

    def _chat_with_websurfer(
        self, message: str, clear_history: bool, **kwargs: Any
    ) -> WebSurferAnswer:
        msg: Optional[str] = message

        while msg is not None:
            chat_result = self.websurfer.initiate_chat(
                self.assistant,
                clear_history=clear_history,
                message=msg,
            )
            msg = self._get_error_message(chat_result)
            clear_history = False

        return self._get_answer(chat_result)

    def _get_error_from_exception(self, task: str, e: Exception) -> str:
        answer = WebSurferAnswer(
            task=task,
            is_successful=False,
            short_answer="unexpected error occurred",
            long_answer=str(e),
            visited_links=[],
        )

        return self.create_final_reply(task, answer)

    def create_final_reply(self, task: str, message: WebSurferAnswer) -> str:
        retval = (
            "We have successfully completed the task:\n\n"
            if message.is_successful
            else "We have failed to complete the task:\n\n"
        )
        retval += f"{task}\n\n"
        retval += f"Short answer: {message.short_answer}\n\n"
        retval += f"Explanation: {message.long_answer}\n\n"
        retval += "Visited links:\n"
        for link in message.visited_links:
            retval += f"  - {link}\n"

        return retval

    def create_new_task(
        self, task: Annotated[str, "a new task for websurfer to perform"]
    ) -> str:
        self.task = task
        try:
            answer = self._chat_with_websurfer(
                message=self.initial_message,
                clear_history=True,
            )
        except Exception as e:
            return self._get_error_from_exception(task, e)

        return self.create_final_reply(task, answer)

    def continue_task_with_additional_instructions(
        self, message: Annotated[str, "a followup message to the existing task"]
    ) -> str:
        try:
            answer = self._chat_with_websurfer(
                message=message,
                clear_history=False,
            )
        except Exception as e:
            return self._get_error_from_exception(message, e)

        return self.create_final_reply(message, answer)

    @property
    def example_answer(self) -> WebSurferAnswer:
        return WebSurferAnswer.get_example_answer()

    @property
    def initial_message(self) -> str:
        return f"""We are tasked with the following task:

{self.task}

If no link is provided in the task, you should search the internet first to find the relevant information.

The focus is on the provided url and its subpages, we do NOT care about the rest of the website i.e. parent pages.
e.g. If the url is 'https://www.example.com/products/air-conditioners', we are interested ONLY in the 'air-conditioners' and its subpages.

AFTER visiting the home page, create a step-by-step plan BEFORE visiting the other pages.
You can click on MAXIMUM {self.max_links_to_click} links. Do NOT try to click all the links on the page, but only the ones which are most relevant for the task (MAX {self.max_links_to_click})!
Do NOT visit the same page multiple times, but only once!
If your co-speaker repeats the same message, inform him that you have already answered to that message and ask him to proceed with the task.
e.g. "I have already answered to that message, please proceed with the task or you will be penalized!"
"""

    @property
    def error_message(self) -> str:
        return f"""Please output the JSON-encoded answer only in the following message before trying to terminate the chat.

IMPORTANT:
  - NEVER enclose JSON-encoded answer in any other text or formatting including '```json' ... '```' or similar!
  - NEVER write TERMINATE in the same message as the JSON-encoded answer!

EXAMPLE:

{self.example_answer.model_dump_json()}

NEGATIVE EXAMPLES:

1. Do NOT include 'TERMINATE' in the same message as the JSON-encoded answer!

{self.example_answer.model_dump_json()}

TERMINATE

2. Do NOT include triple backticks or similar!

```json
{self.example_answer.model_dump_json()}
```

THE LAST ERROR MESSAGE:

{self.last_is_termination_msg_error}

"""

    @property
    def system_message(self) -> str:
        return f"""You are in charge of navigating the web_surfer agent to scrape the web.
web_surfer is able to CLICK on links, SCROLL down, and scrape the content of the web page. e.g. you cen tell him: "Click the 'Getting Started' result".
Each time you receive a reply from web_surfer, you need to tell him what to do next. e.g. "Click the TV link" or "Scroll down".
It is very important that you explore ONLY the page links relevant for the task!

GUIDELINES:
- Once you retrieve the content from the received url, you can tell web_surfer to CLICK on links, SCROLL down...
By using these capabilities, you will be able to retrieve MUCH BETTER information from the web page than by just scraping the given URL!
You MUST use these capabilities when you receive a task for a specific category/product etc.
- do NOT try to create a summary without clicking on any link, because you will be missing a lot of information!
- if needed, you can instruct web surfer to SEARCH THE WEB for information.

Examples:
"Click the 'TVs' result" - This way you will navigate to the TVs section of the page and you will find more information about TVs.
"Click 'Electronics' link" - This way you will navigate to the Electronics section of the page and you will find more information about Electronics.
"Click the 'Next' button"
"Search the internet for the best TV to buy" - this will get links to initial pages to start the search

- Do NOT try to click all the links on the page, but only the ones which are RELEVANT for the task! Web pages can be very long and you will be penalized if spend too much time on this task!
- Your final goal is to summarize the findings for the given task. The summary must be in English!
- Create a summary after you successfully retrieve the information from the web page.
- It is useful to include in the summary relevant links where more information can be found.
e.g. If the page is offering to sell TVs, you can include a link to the TV section of the page.
- If you get some 40x error, please do NOT give up immediately, but try to navigate to another page and continue with the task.
Give up only if you get 40x error on ALL the pages which you tried to navigate to.


FINAL MESSAGE:
Once you have retrieved he wanted information, YOU MUST create JSON-encoded string. Summary created by the web_surfer is not enough!
You MUST not include any other text or formatting in the message, only JSON-encoded summary!

An example of the JSON-encoded summary:
{self.example_answer.model_dump_json()}

TERMINATION:
When YOU are finished and YOU have created JSON-encoded answer, write a single 'TERMINATE' to end the task.

OFTEN MISTAKES:
- Web surfer expects you to tell him what LINK NAME to click next, not the relative link. E.g. in case of '[Hardware](/Hardware), the proper command would be 'Click into 'Hardware''.
- Links presented are often RELATIVE links, so you need to ADD the DOMAIN to the link to make it work. E.g. link '/products/air-conditioners' should be 'https://www.example.com/products/air-conditioners'
- You do NOT need to click on MAX number of links. If you have enough information from the first xy links, you do NOT need to click on the rest of the links!
- Do NOT repeat the steps you have already completed!
- ALWAYS include the NEXT steps in the message!
- Do NOT instruct web_surfer to click on the same link multiple times. If there are some problems with the link, MOVE ON to the next one!
- Also, if web_surfer does not understand your message, just MOVE ON to the next link!
- NEVER REPEAT the same instructions to web_surfer! If he does not understand the first time, MOVE ON to the next link!
- NEVER enclose JSON-encoded answer in any other text or formatting including '```json' ... '```' or similar!
"""

    def register(
        self,
        *,
        caller: AutoGenConversableAgent,
        executor: Union[AutoGenConversableAgent, list[AutoGenConversableAgent]],
    ) -> None:
        @caller.register_for_llm(  # type: ignore[misc]
            name="create_new_websurfing_task",
            description="Creates a new task for a websurfer that can include searching or browsing the internet.",
        )
        def create_new_task(
            task: Annotated[str, "a new task for websurfer to perform"],
        ) -> str:
            return self.create_new_task(task)

        @caller.register_for_llm(  # type: ignore[misc]
            name="continue_websurfing_task_with_additional_instructions",
            description="Continue an existing task for a websurfer with additional instructions.",
        )
        def continue_task_with_additional_instructions(
            message: Annotated[
                str,
                "Additional instructions for the task after receiving the initial answer",
            ],
        ) -> str:
            return self.continue_task_with_additional_instructions(message)

        executors = executor if isinstance(executor, list) else [executor]
        for executor in executors:
            executor.register_for_execution(name="create_new_websurfing_task")(
                create_new_task
            )
            executor.register_for_execution(
                name="continue_websurfing_task_with_additional_instructions"
            )(continue_task_with_additional_instructions)



__all__ = ["FastAgency"]

from collections.abc import Awaitable, Generator
from contextlib import contextmanager
from typing import Any, Callable, Optional, Union

from .base import ASGI, UI, WSGI, Workflows
from .exceptions import (
    FastAgencyASGINotImplementedError,
    FastAgencyWSGINotImplementedError,
)
from .logging import get_logger

logger = get_logger(__name__)


class FastAgency:  # Runnable
    def __init__(
        self,
        wf: Workflows,
        ui: UI,
        *,
        title: Optional[str] = None,
        description: Optional[str] = None,
    ) -> None:
        """Initialize the FastAgency object.

        Args:
            wf (Workflows): The workflows object to use
            ui (UI): The UI object to use
            title (Optional[str], optional): The title of the FastAgency. If None, the default string will be used. Defaults to None.
            description (Optional[str], optional): The description of the FastAgency. If None, the default string will be used. Defaults to None.
        """
        self._title = title or "FastAgency application"
        default_description = "FastAgency application"

        if len(wf.names) == 0:
            logger.warning(f"No workflows found in {wf}")
            default_description += " - No workflows found"
        else:
            default_description += " - Workflows:"
            for name in wf.names:
                default_description += f" - {name}: {wf.get_description(name)}"
        self._description = description or default_description

        logger.info(f"Initializing FastAgency {self} with workflows: {wf} and UI: {ui}")
        self._wf = wf
        self._ui = ui
        logger.info(f"Initialized FastAgency: {self}")

    @property
    def title(self) -> str:
        """Return the title of the FastAgency."""
        return self._title

    @property
    def description(self) -> str:
        """Return the description of the FastAgency."""
        return self._description

    def __str__(self) -> str:
        """Return the string representation of the FastAgency."""
        return f"<FastAgency title={self._title}>"

    @property
    def wf(self) -> Workflows:
        """Return the workflows object."""
        return self._wf

    @property
    def ui(self) -> UI:
        """Return the UI object."""
        return self._ui

    @contextmanager
    def create(self, import_string: str) -> Generator[None, None, None]:
        """Create the FastAgency."""
        with self._ui.create(app=self, import_string=import_string):
            yield

    def start(
        self,
        import_string: str,
        name: Optional[str] = None,
        initial_message: Optional[str] = None,
        single_run: bool = False,
    ) -> None:
        """Start the FastAgency."""
        self.ui.start(
            app=self,
            import_string=import_string,
            name=name,
            initial_message=initial_message,
            single_run=single_run,
        )

    def __call__(self, *args: Any) -> Union[Awaitable[None], list[bytes]]:
        if len(args) == 2 and callable(args[1]):
            # WSGI interface
            environ, start_response = args
            return self.handle_wsgi(environ, start_response)
        elif len(args) == 3 and callable(args[1]) and callable(args[2]):
            # ASGI interface
            scope, receive, send = args
            scope_type = scope.get("type")
            if scope_type == "http":
                return self.handle_asgi(scope, receive, send)
            else:
                raise NotImplementedError(
                    f"ASGI scope type '{scope_type}' not supported."
                )
        else:
            raise TypeError(f"Invalid arguments for __call__: {args}")

    def handle_wsgi(
        self, environ: dict[str, Any], start_response: Callable[..., Any]
    ) -> list[bytes]:
        if isinstance(self.ui, WSGI):
            return self.ui.handle_wsgi(self, environ, start_response)
        else:
            raise FastAgencyWSGINotImplementedError(
                "WSGI interface not supported for UI: {self.ui}"
            )

    async def handle_asgi(
        self,
        scope: dict[str, Any],
        receive: Callable[[dict[str, Any]], Awaitable[None]],
        send: Callable[[dict[str, Any]], Awaitable[None]],
    ) -> None:
        if isinstance(self.ui, ASGI):
            return await self.ui.handle_asgi(self, scope, receive, send)
        else:
            raise FastAgencyASGINotImplementedError(
                "ASGI interface not supported for UI: {self.ui}"
            )



import logging
from typing import Any, ClassVar, Literal, Optional, Protocol

from pydantic import BaseModel, model_validator

# Get the logger
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class BaseSecurity(BaseModel):
    """Base class for security classes."""

    type: ClassVar[Literal["apiKey", "http", "mutualTLS", "oauth2", "openIdConnect"]]
    in_value: ClassVar[Literal["header", "query", "cookie", "bearer", "basic", "tls"]]
    name: str

    @model_validator(mode="after")  # type: ignore[misc]
    def __post_init__(
        self,
    ) -> None:  # dataclasses uses __post_init__ instead of model_validator
        """Validate the in_value based on the type."""
        valid_in_values = {
            "apiKey": ["header", "query", "cookie"],
            "http": ["bearer", "basic"],
            "oauth2": ["bearer"],
            "openIdConnect": ["bearer"],
            "mutualTLS": ["tls"],
        }
        if self.in_value not in valid_in_values[self.type]:
            raise ValueError(
                f"Invalid in_value '{self.in_value}' for type '{self.type}'"
            )

    def accept(self, security_params: "BaseSecurityParameters") -> bool:
        return isinstance(self, security_params.get_security_class())

    @classmethod
    def is_supported(cls, type: str, in_value: str) -> bool:
        return type == cls.type and in_value == cls.in_value

    @classmethod
    def get_security_class(cls, type: str, in_value: str) -> Optional[str]:
        sub_classes = cls.__subclasses__()

        for sub_class in sub_classes:
            if sub_class.is_supported(type, in_value):
                return sub_class.__name__
        else:
            logger.error(
                f"Unsupported type '{type}' and in_value '{in_value}' combination"
            )
            return None


class BaseSecurityParameters(Protocol):
    """Base class for security parameters."""

    def apply(
        self,
        q_params: dict[str, Any],
        body_dict: dict[str, Any],
        security: BaseSecurity,
    ) -> None: ...

    def get_security_class(self) -> type[BaseSecurity]: ...


class APIKeyHeader(BaseSecurity):
    """API Key Header security class."""

    type: ClassVar[Literal["apiKey"]] = "apiKey"
    in_value: ClassVar[Literal["header"]] = "header"

    class Parameters(BaseModel):  # BaseSecurityParameters
        """API Key Header security parameters class."""

        value: str

        def apply(
            self,
            q_params: dict[str, Any],
            body_dict: dict[str, Any],
            security: BaseSecurity,
        ) -> None:
            api_key_header: APIKeyHeader = security  # type: ignore[assignment]

            if "headers" not in body_dict:
                body_dict["headers"] = {}

            body_dict["headers"][api_key_header.name] = self.value

        def get_security_class(self) -> type[BaseSecurity]:
            return APIKeyHeader


class APIKeyQuery(BaseSecurity):
    """API Key Query security class."""

    type: ClassVar[Literal["apiKey"]] = "apiKey"
    in_value: ClassVar[Literal["query"]] = "query"

    class Parameters(BaseModel):  # BaseSecurityParameters
        """API Key Query security parameters class."""

        value: str

        def apply(
            self,
            q_params: dict[str, Any],
            body_dict: dict[str, Any],
            security: BaseSecurity,
        ) -> None:
            api_key_query: APIKeyQuery = security  # type: ignore[assignment]

            q_params[api_key_query.name] = self.value

        def get_security_class(self) -> type[BaseSecurity]:
            return APIKeyQuery


class APIKeyCookie(BaseSecurity):
    """API Key Cookie security class."""

    type: ClassVar[Literal["apiKey"]] = "apiKey"
    in_value: ClassVar[Literal["cookie"]] = "cookie"

    class Parameters(BaseModel):  # BaseSecurityParameters
        """API Key Cookie security parameters class."""

        value: str

        def apply(
            self,
            q_params: dict[str, Any],
            body_dict: dict[str, Any],
            security: BaseSecurity,
        ) -> None:
            api_key_cookie: APIKeyCookie = security  # type: ignore[assignment]

            if "cookies" not in body_dict:
                body_dict["cookies"] = {}

            body_dict["cookies"][api_key_cookie.name] = self.value

        def get_security_class(self) -> type[BaseSecurity]:
            return APIKeyCookie


class HTTPBearer(BaseSecurity):
    """HTTP Bearer security class."""

    type: ClassVar[Literal["http"]] = "http"
    in_value: ClassVar[Literal["bearer"]] = "bearer"

    class Parameters(BaseModel):  # BaseSecurityParameters
        """HTTP Bearer security parameters class."""

        value: str

        def apply(
            self,
            q_params: dict[str, Any],
            body_dict: dict[str, Any],
            security: BaseSecurity,
        ) -> None:
            if "headers" not in body_dict:
                body_dict["headers"] = {}

            body_dict["headers"]["Authorization"] = f"Bearer {self.value}"

        def get_security_class(self) -> type[BaseSecurity]:
            return HTTPBearer



class FastAgencyError(Exception):
    pass


class FastAgencyCLIError(FastAgencyError):
    pass


class FastAgencyCLIPythonVersionError(FastAgencyCLIError):
    pass


class FastAgencyWSGINotImplementedError(FastAgencyError):
    pass


class FastAgencyASGINotImplementedError(FastAgencyError):
    pass



import json
from collections.abc import Iterable, Iterator
from typing import Any, Callable, Optional
from uuid import uuid4

import mesop as me

from ...base import (
    AskingMessage,
    FunctionCallExecution,
    IOMessage,
    IOMessageVisitor,
    MultipleChoice,
    SuggestedFunctionCall,
    SystemMessage,
    TextInput,
    TextMessage,
    WorkflowCompleted,
)
from ...logging import get_logger
from .base import MesopMessage
from .components.inputs import input_text
from .data_model import Conversation, ConversationMessage, State
from .send_prompt import send_user_feedback_to_autogen
from .styles import MesopHomePageStyles, MesopMessageStyles

logger = get_logger(__name__)


def dict_to_markdown(d: dict[str, Any], *, indent: int = 2) -> str:
    return "\n```\n" + json.dumps(d, indent=2) + "\n```\n"


def consume_responses(responses: Iterable[MesopMessage]) -> Iterator[None]:
    for message in responses:
        state = me.state(State)
        handle_message(state, message)
        yield
        me.scroll_into_view(key="end_of_messages")
        yield
    yield


def handle_message(state: State, message: MesopMessage) -> None:
    conversation = state.conversation
    messages = conversation.messages
    level = message.conversation.level
    conversation_id = message.conversation.id
    io_message = message.io_message
    message_dict = io_message.model_dump()
    message_json = json.dumps(message_dict)
    conversation_message = ConversationMessage(
        level=level,
        conversation_id=conversation_id,
        io_message_json=message_json,
        feedback=[],
    )
    messages.append(conversation_message)
    conversation.messages = list(messages)
    if isinstance(io_message, AskingMessage):
        conversation.waiting_for_feedback = True
        conversation.completed = False
    if isinstance(io_message, WorkflowCompleted):
        conversation.completed = True
        conversation.waiting_for_feedback = False
        if not conversation.is_from_the_past:
            uuid: str = uuid4().hex
            becomme_past = Conversation(
                id=uuid,
                title=conversation.title,
                messages=conversation.messages,
                completed=True,
                is_from_the_past=True,
                waiting_for_feedback=False,
            )
            state.past_conversations.insert(0, becomme_past)


def message_box(
    message: ConversationMessage, read_only: bool, *, styles: MesopHomePageStyles
) -> None:
    io_message_dict = json.loads(message.io_message_json)
    level = message.level
    conversation_id = message.conversation_id
    io_message = IOMessage.create(**io_message_dict)
    visitor = MesopGUIMessageVisitor(level, conversation_id, message, styles, read_only)
    visitor.process_message(io_message)


class MesopGUIMessageVisitor(IOMessageVisitor):
    def __init__(
        self,
        level: int,
        conversation_id: str,
        conversation_message: ConversationMessage,
        styles: MesopHomePageStyles,
        read_only: bool = False,
    ) -> None:
        """Initialize the MesopGUIMessageVisitor object.

        Args:
            level (int): The level of the message.
            conversation_id (str): The ID of the conversation.
            conversation_message (ConversationMessage): Conversation message that wraps the visited io_message
            styles (MesopHomePageStyles): Styles for the message
            read_only (bool): Input messages are disabled in read only mode
        """
        self._level = level
        self._conversation_id = conversation_id
        self._readonly = read_only
        self._conversation_message = conversation_message
        self._styles = styles

    def _has_feedback(self) -> bool:
        return len(self._conversation_message.feedback) > 0

    def _is_completed(self) -> bool:
        return self._conversation_message.feedback_completed

    def _provide_feedback(self, feedback: str) -> Iterator[None]:
        state = me.state(State)
        conversation = state.conversation
        conversation.feedback = ""
        conversation.waiting_for_feedback = False
        yield
        me.scroll_into_view(key="end_of_messages")
        yield
        responses = send_user_feedback_to_autogen(feedback)
        yield from consume_responses(responses)

    def message_content_to_markdown(self, msg: IOMessage) -> str:
        d = msg.model_dump()
        d.pop("sender")
        d.pop("recipient")
        d.pop("type")

        return "\n".join([f"**{k}**: {v} <br>" for k, v in d["content"].items()])

    def visit_default(
        self,
        message: IOMessage,
        *,
        content: Optional[str] = None,
        style: Optional[MesopMessageStyles] = None,
        error: Optional[bool] = False,
        inner_callback: Optional[Callable[..., None]] = None,
    ) -> None:
        logger.info(f"visit_default: {message=}")
        style = style or self._styles.message.default
        title = message.type.replace("_", " ").capitalize()
        title = "[Error] " + title if error else title
        with me.box(style=style.box or self._styles.message.default.box):
            self._header(
                message,
                title=title,
                box_style=style.header_box,
                md_style=style.header_md,
            )

            content = content or self.message_content_to_markdown(message)

            me.markdown(content, style=style.md or self._styles.message.default.md)

            if inner_callback:
                inner_callback()

    def visit_text_message(self, message: TextMessage) -> None:
        content = message.body if message.body else ""
        content = content if content.strip() != "" else "*(empty message)*"
        self.visit_default(
            message,
            content=content,
            style=self._styles.message.text,
        )

    def visit_system_message(self, message: SystemMessage) -> None:
        content = (
            f"""#### **{message.message['heading']}**

{message.message['body']}
"""
            if "heading" in message.message and "body" in message.message
            else None
        )

        self.visit_default(
            message,
            content=content,
            style=self._styles.message.system,
        )

    def visit_suggested_function_call(self, message: SuggestedFunctionCall) -> None:
        content = f"""
**function_name**: `{message.function_name}`<br>
**call_id**: `{message.call_id}`<br>
**arguments**:
{dict_to_markdown(message.arguments)}
"""
        self.visit_default(
            message,
            content=content,
            style=self._styles.message.suggested_function_call,
        )

    def visit_function_call_execution(self, message: FunctionCallExecution) -> None:
        return self.visit_default(
            message,
            style=self._styles.message.function_call_execution,
        )

    def visit_text_input(self, message: TextInput) -> str:
        def on_input(feedback: str) -> Iterator[None]:
            self._conversation_message.feedback = [feedback]
            self._conversation_message.feedback_completed = True
            yield from self._provide_feedback(feedback)

        def value_if_completed() -> Optional[str]:
            message = self._conversation_message
            return message.feedback[0] if message.feedback_completed else None

        # base_color = "#dff"
        prompt = message.prompt if message.prompt else "Please enter a value"
        if message.suggestions:
            suggestions = ",".join(suggestion for suggestion in message.suggestions)
            prompt += "\n Suggestions: " + suggestions

        self.visit_default(
            message,
            content=prompt,
            style=self._styles.message.text_input,
            inner_callback=lambda: input_text(
                on_input,
                key="prompt",
                disabled=self._readonly or self._has_feedback(),
                value=value_if_completed(),
                style=self._styles.message.text_input_inner,
            ),
        )
        return ""

    def visit_multiple_choice(self, message: MultipleChoice) -> str:
        if message.single:
            return self._visit_single_choice(message)
        else:
            return self._visit_many_choices(message)

    def _visit_single_choice(self, message: MultipleChoice) -> str:
        def on_change(ev: me.RadioChangeEvent) -> Iterator[None]:
            feedback = ev.value
            self._conversation_message.feedback = [feedback]
            self._conversation_message.feedback_completed = True
            yield from self._provide_feedback(feedback)

        # base_color = "#dff"
        prompt = message.prompt if message.prompt else "Please enter a value"
        if message.choices:
            options = (
                me.RadioOption(
                    label=(choice if choice != message.default else choice + " *"),
                    value=choice,
                )
                for choice in message.choices
            )
        if self._has_feedback():
            pre_selected = {"value": self._conversation_message.feedback[0]}
        else:
            pre_selected = {}

        def inner_callback() -> None:
            me.radio(
                on_change=on_change,
                disabled=self._readonly or self._is_completed(),
                options=options,
                style=self._styles.message.single_choice_inner.radio,
                **pre_selected,
            )

        self.visit_default(
            message,
            content=prompt,
            style=self._styles.message.text_input,
            inner_callback=inner_callback,
        )
        return ""

    def _visit_many_choices(self, message: MultipleChoice) -> str:
        def on_change(ev: me.CheckboxChangeEvent) -> None:
            message_feedback = self._conversation_message.feedback
            choice = ev.key
            yes_no = ev.checked
            if yes_no:
                if choice not in message_feedback:
                    message_feedback.append(choice)
            else:
                if choice in message_feedback:
                    message_feedback.remove(choice)

        def on_click(ev: me.ClickEvent) -> Iterator[None]:
            message_feedback = self._conversation_message.feedback
            feedback = ",".join(message_feedback)
            self._conversation_message.feedback_completed = True
            yield from self._provide_feedback(feedback)

        def should_be_checked(option: str) -> bool:
            conversation_message = self._conversation_message
            return option in conversation_message.feedback

        prompt = message.prompt if message.prompt else "Please select a value:"

        def inner_callback() -> None:
            if message.choices:
                with me.box(
                    style=self._styles.message.multiple_choice_inner.box,
                ):
                    for option in message.choices:
                        me.checkbox(
                            label=option,
                            key=option,
                            checked=should_be_checked(option),
                            on_change=on_change,
                            disabled=self._readonly or self._is_completed(),
                            style=self._styles.message.multiple_choice_inner.checkbox,
                        )
                    me.button(
                        label="OK",
                        on_click=on_click,
                        color="primary",
                        type="flat",
                        style=self._styles.message.multiple_choice_inner.button,
                    )

        self.visit_default(
            message,
            content=prompt,
            style=self._styles.message.text_input,
            inner_callback=inner_callback,
        )
        return ""

    def render_error_message(
        self,
        e: Exception,
        message: IOMessage,
        *,
        content: Optional[str] = None,
        style: Optional[MesopMessageStyles] = None,
    ) -> None:
        style = self._styles.message.error or self._styles.message.default
        title = "[Error] " + message.type.replace("_", " ").capitalize()

        with me.box(style=style.box or self._styles.message.default.box):
            self._header(
                message,
                title=title,
                box_style=style.header_box or self._styles.message.default.header_box,
                md_style=style.header_md or self._styles.message.default.header_md,
            )

            content = (
                "Failed to render message:"
                + dict_to_markdown(message.model_dump())
                + f"<br>Error: {e}"
            )

            logger.info(f"render_error_message: {content=}")
            me.markdown(content, style=style.md or self._styles.message.default.md)

    def process_message(self, message: IOMessage) -> Optional[str]:
        try:
            return self.visit(message)
        except Exception as e:
            logger.warning(f"Failed to render message: {e}")
            self.render_error_message(e, message)
            return None

    def _header(
        self,
        message: IOMessage,
        *,
        title: Optional[str] = None,
        box_style: Optional[me.Style] = None,
        md_style: Optional[me.Style] = None,
    ) -> None:
        with me.box(style=box_style or self._styles.message.default.header_box):
            h = title if title else message.type
            if message.sender and message.recipient:
                h += f": {message.sender} -> {message.recipient}"
            elif message.sender:
                h += f": to {message.sender}"
            elif message.recipient:
                h += f": from {message.recipient}"
            if message.auto_reply:
                h += " (auto-reply)"
            h = f"**{h}**"
            # style=me.Style(padding=me.Padding(top=8, right=16, left=16, bottom=8))
            me.markdown(h, style=md_style or self._styles.message.default.header_md)



from dataclasses import dataclass, field

import mesop as me

from .components.helpers import darken_hex_color

HEADER_BOX_STYLE = me.Style(
    padding=me.Padding(bottom="24px"),
)

HEADER_TEXT_STYLE = me.Style(
    font_weight=500,
    font_size=24,
    color="#3D3929",
    letter_spacing="0.3px",
)


ROOT_BOX_STYLE = me.Style(
    background="#e7f2ff",
    height="100%",
    font_family="Inter",
    display="flex",
    flex_direction="row",
)

PAST_CHATS_SHOW_STYLE = me.Style(
    background=darken_hex_color("#e7f2ff", 0.98),
    height="100%",
    width="min(300px)",
    font_family="Inter",
    display="flex",
    flex_direction="column",
)

PAST_CHATS_HIDE_STYLE = me.Style(
    background="#e7f2ff",
    height="10%",
    width="min(150px)",
    font_family="Inter",
    display="flex",
    flex_direction="column",
)

PAST_CHATS_INNER_STYLE = me.Style(
    flex_direction="row",
    width="100%",
    padding=me.Padding(top="16px"),
    justify_content="space-between",
)

PAST_CHATS_CONV_STYLE = me.Style(
    padding=me.Padding.all(16),
    border_radius=16,
)


CHAT_STARTER_STYLE = me.Style(
    background="#e7f2ff",
    height="100%",
    width="max(80%)",
    font_family="Inter",
    display="flex",
    flex_direction="column",
    padding=me.Padding.all(16),
)

CONV_STARTER_STYLE = me.Style(
    width="min(680px, 100%)",
    # margin=me.Margin.symmetric(horizontal="auto", vertical=36),
)

CONV_STARTER_TEXT_STYLE = me.Style(font_size=20, margin=me.Margin(bottom=24))

CONV_LIST_STYLE = me.Style(
    overflow_y="auto",
    display="flex",
    flex_direction="column",
    padding=me.Padding.symmetric(horizontal=16),
)

CONV_TOP_STYLE = me.Style(
    margin=me.Margin(bottom="1vh"),
)

CONV_MSG_STYLE = me.Style(margin=me.Margin(bottom="50vh"))

STYLESHEETS = [
    "https://fonts.googleapis.com/css2?family=Inter:wght@100..900&display=swap"
]

MSG_DEFAULT_HEADER_MD_STYLE = me.Style(
    padding=me.Padding(top=8, right=16, left=16, bottom=8)
)
MSG_DEFAULT_HEADER_BOX_STYLE = me.Style(
    background="#ddd",
    padding=me.Padding.all(0),
    border_radius=8,
)

MSG_DEFAULT_BOX_STYLE = me.Style(
    background="#fff",
    border_radius=8,
    margin=me.Margin.symmetric(vertical=16),
)

MSG_DEFAULT_MD_STYLE = me.Style(padding=me.Padding.all(16))

MSG_SYSTEM_HEADER_BOX_STYLE = me.Style(
    background="#b3e5fc",
    padding=me.Padding.all(0),
    border_radius=8,
)

MSG_SUGESTED_FUNCTION_CALL_HEADER_BOX_STYLE = me.Style(
    background="#fff0b2",
    padding=me.Padding.all(0),
    border_radius=8,
)

MSG_FUNCTION_CALL_EXECUTION_HEADER_BOX_STYLE = me.Style(
    background="#ffccbc",
    padding=me.Padding.all(0),
    border_radius=8,
)

MSG_ERROR_HEADER_BOX_STYLE = me.Style(
    background="#f44336",
    padding=me.Padding.all(0),
    border_radius=8,
)

MSG_TEXT_INPUT_HEADER_BOX_STYLE = me.Style(
    background="#e6ee9c",
    padding=me.Padding.all(0),
    border_radius=8,
)

MSG_TEXT_INPUT_MD_STYLE = me.Style(
    background="#e0e0e0",
    padding=me.Padding.all(16),
)


@dataclass
class MesopMessageStyles:
    box: me.Style = field(default_factory=lambda: MSG_DEFAULT_BOX_STYLE)
    md: me.Style = field(default_factory=lambda: MSG_DEFAULT_MD_STYLE)
    header_box: me.Style = field(default_factory=lambda: MSG_DEFAULT_HEADER_BOX_STYLE)
    header_md: me.Style = field(default_factory=lambda: MSG_DEFAULT_HEADER_MD_STYLE)


TEXT_INPUT_INNER_BOX_STYLE = me.Style(
    border_radius=8,
    padding=me.Padding.all(16),
    background="white",
    display="flex",
    width="100%",
)

TEXT_INPUT_NATIVE_TEXTAREA_STYLE = me.Style(
    padding=me.Padding(top=16, left=16),
    # outline="none",
    width="100%",
    border=me.Border.all(me.BorderSide(style="none")),
    background="#e1f5fe",
)

SINGLE_CHOICE_RADIO_STYLE = me.Style(
    display="flex", flex_direction="column", padding=me.Padding.all(16)
)

MULTIPLE_CHOICE_RADIO_STYLE = me.Style(
    display="flex",
    flex_direction="column",
    padding=me.Padding.all(16),
)

MULTIPLE_CHOICE_BUTTON_STYLE = me.Style(
    margin=me.Margin(top=16), border_radius=8, background="#1976d2"
)

MULTIPLE_CHOICE_CHECKBOX_STYLE = me.Style()


@dataclass
class MesopTextInputInnerStyles:
    box: me.Style = field(default_factory=lambda: TEXT_INPUT_INNER_BOX_STYLE)
    native_textarea: me.Style = field(
        default_factory=lambda: TEXT_INPUT_NATIVE_TEXTAREA_STYLE
    )


@dataclass
class MesopSingleChoiceInnerStyles:
    radio: me.Style = field(default_factory=lambda: SINGLE_CHOICE_RADIO_STYLE)


@dataclass
class MesopMultipleChoiceInnerStyles:
    box: me.Style = field(default_factory=lambda: MULTIPLE_CHOICE_RADIO_STYLE)
    checkbox: me.Style = field(default_factory=lambda: MULTIPLE_CHOICE_CHECKBOX_STYLE)
    button: me.Style = field(default_factory=lambda: MULTIPLE_CHOICE_BUTTON_STYLE)


@dataclass
class MesopMessagesStyles:
    default: MesopMessageStyles = field(default_factory=lambda: MesopMessageStyles())

    error: MesopMessageStyles = field(
        default_factory=lambda: MesopMessageStyles(
            header_box=MSG_ERROR_HEADER_BOX_STYLE,
        )
    )

    system: MesopMessageStyles = field(
        default_factory=lambda: MesopMessageStyles(
            header_box=MSG_SYSTEM_HEADER_BOX_STYLE,
        )
    )

    text: MesopMessageStyles = field(default_factory=lambda: MesopMessageStyles())

    suggested_function_call: MesopMessageStyles = field(
        default_factory=lambda: MesopMessageStyles(
            header_box=MSG_SUGESTED_FUNCTION_CALL_HEADER_BOX_STYLE,
        )
    )

    function_call_execution: MesopMessageStyles = field(
        default_factory=lambda: MesopMessageStyles(
            header_box=MSG_FUNCTION_CALL_EXECUTION_HEADER_BOX_STYLE,
        )
    )

    text_input: MesopMessageStyles = field(
        default_factory=lambda: MesopMessageStyles(
            header_box=MSG_TEXT_INPUT_HEADER_BOX_STYLE,
        )
    )

    text_input_inner: MesopTextInputInnerStyles = field(
        default_factory=lambda: MesopTextInputInnerStyles()
    )

    single_choice_inner: MesopSingleChoiceInnerStyles = field(
        default_factory=lambda: MesopSingleChoiceInnerStyles()
    )

    multiple_choice_inner: MesopMultipleChoiceInnerStyles = field(
        default_factory=lambda: MesopMultipleChoiceInnerStyles()
    )


@dataclass
class MesopHomePageStyles:
    chat_starter: me.Style = field(default_factory=lambda: CHAT_STARTER_STYLE)
    conv_list: me.Style = field(default_factory=lambda: CONV_LIST_STYLE)
    conv_msg: me.Style = field(default_factory=lambda: CONV_MSG_STYLE)
    conv_top: me.Style = field(default_factory=lambda: CONV_TOP_STYLE)
    conv_starter: me.Style = field(default_factory=lambda: CONV_STARTER_STYLE)
    conv_starter_text: me.Style = field(default_factory=lambda: CONV_STARTER_TEXT_STYLE)
    header: me.Style = field(default_factory=lambda: HEADER_BOX_STYLE)
    header_text: me.Style = field(default_factory=lambda: HEADER_TEXT_STYLE)
    past_chats_hide: me.Style = field(default_factory=lambda: PAST_CHATS_HIDE_STYLE)
    past_chats_show: me.Style = field(default_factory=lambda: PAST_CHATS_SHOW_STYLE)
    past_chats_inner: me.Style = field(default_factory=lambda: PAST_CHATS_INNER_STYLE)
    past_chats_conv: me.Style = field(default_factory=lambda: PAST_CHATS_CONV_STYLE)
    root: me.Style = field(default_factory=lambda: ROOT_BOX_STYLE)
    stylesheets: list[str] = field(default_factory=lambda: STYLESHEETS)
    message: MesopMessagesStyles = field(default_factory=lambda: MesopMessagesStyles())



import re
import textwrap
from abc import ABC, abstractmethod
from collections.abc import Awaitable, Generator, Iterable, Iterator, Mapping
from contextlib import contextmanager
from dataclasses import asdict, dataclass, field, fields
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Literal,
    Optional,
    Protocol,
    Type,
    TypeVar,
    Union,
    runtime_checkable,
)

if TYPE_CHECKING:
    from fastagency.api.openapi import OpenAPI

__all__ = [
    "UI",
    "FunctionCallExecution",
    "IOMessage",
    "MessageType",
    "MultipleChoice",
    "Runnable",
    "SuggestedFunctionCall",
    "SystemMessage",
    "TextInput",
    "TextMessage",
    "Workflow",
    "WorkflowCompleted",
    "Workflows",
    "run_workflow",
]

MessageType = Literal[
    "text_message",
    "suggested_function_call",
    "function_call_execution",
    "text_input",
    "multiple_choice",
    "system_message",
    "workflow_completed",
]


def _camel_to_snake(name: str) -> str:
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


@dataclass
class IOMessage(ABC):  # noqa: B024  # `IOMessage` is an abstract base class, but it has no abstract methods
    sender: Optional[str] = None
    recipient: Optional[str] = None
    # streaming: bool = False
    auto_reply: bool = False

    @property
    def type(self) -> MessageType:
        retval: MessageType = _camel_to_snake(self.__class__.__name__)  # type: ignore[assignment]
        return retval

    @staticmethod
    def _get_message_class(type: Optional[MessageType]) -> "Type[IOMessage]":
        type = type or "text_message"
        lookup: dict[MessageType, "Type[IOMessage]"] = {
            "text_message": TextMessage,
            "suggested_function_call": SuggestedFunctionCall,
            "function_call_execution": FunctionCallExecution,
            "text_input": TextInput,
            "multiple_choice": MultipleChoice,
            "system_message": SystemMessage,
            "workflow_completed": WorkflowCompleted,
        }
        return lookup[type]

    @staticmethod
    def create(type: Optional[MessageType] = None, **kwargs: Any) -> "IOMessage":
        cls = IOMessage._get_message_class(type)

        content = kwargs.pop("content", {})
        kwargs.update(content)

        return cls(**kwargs)

    @staticmethod
    def _get_parameters_names() -> list[str]:
        return [field.name for field in fields(IOMessage)]

    def model_dump(self) -> dict[str, Any]:
        params_names = IOMessage._get_parameters_names()
        d = asdict(self)
        content = {k: v for k, v in d.items() if k not in params_names}
        retval = {k: v for k, v in d.items() if k in params_names}
        retval["content"] = content
        retval["type"] = self.type
        return retval


# message type that asks user something


@dataclass
class AskingMessage(IOMessage): ...


# type of output messages
@dataclass
class TextMessage(IOMessage):
    body: Optional[str] = None

    def __post_init__(self) -> None:
        """Set the default value for the `type` attribute."""
        if self.type is None:
            self.type = "text_message"


@dataclass
class SuggestedFunctionCall(IOMessage):
    function_name: Optional[str] = None
    call_id: Optional[str] = None
    arguments: dict[str, Any] = field(default_factory=dict)


@dataclass
class FunctionCallExecution(IOMessage):
    function_name: Optional[str] = None
    call_id: Optional[str] = None
    retval: Any = None


# types of input messages
@dataclass
class TextInput(AskingMessage):
    prompt: Optional[str] = None
    suggestions: list[str] = field(default_factory=list)
    password: bool = False


@dataclass
class MultipleChoice(AskingMessage):
    prompt: Optional[str] = None
    choices: list[str] = field(default_factory=list)
    default: Optional[str] = None
    single: bool = True
    # todo: add validation


@dataclass
class SystemMessage(IOMessage):
    message: dict[str, Any] = field(default_factory=dict)


@dataclass
class WorkflowCompleted(IOMessage):
    result: Optional[str] = None


class IOMessageVisitor(ABC):
    def visit(self, message: IOMessage) -> Optional[str]:
        method_name = f"visit_{message.type}"
        method = getattr(self, method_name, self.visit_default)
        return method(message)

    @abstractmethod
    def visit_default(self, message: IOMessage) -> Optional[str]: ...

    def visit_text_message(self, message: TextMessage) -> Optional[str]:
        return self.visit_default(message)

    def visit_suggested_function_call(
        self, message: SuggestedFunctionCall
    ) -> Optional[str]:
        return self.visit_default(message)

    def visit_function_call_execution(
        self, message: FunctionCallExecution
    ) -> Optional[str]:
        return self.visit_default(message)

    def visit_text_input(self, message: TextInput) -> Optional[str]:
        return self.visit_default(message)

    def visit_multiple_choice(self, message: MultipleChoice) -> Optional[str]:
        return self.visit_default(message)

    def visit_system_message(self, message: SystemMessage) -> Optional[str]:
        return self.visit_default(message)

    def visit_workflow_completed(self, message: WorkflowCompleted) -> Optional[str]:
        return self.visit_default(message)


# @dataclass
# class IOStreamingMessage:
#     chunk: str


@runtime_checkable
class UI(Protocol):
    @contextmanager
    def create(self, app: "Runnable", import_string: str) -> Iterator[None]: ...

    def start(
        self,
        *,
        app: "Runnable",
        import_string: str,
        name: Optional[str] = None,
        initial_message: Optional[str] = None,
        single_run: bool = False,
    ) -> None: ...

    def process_message(self, message: IOMessage) -> Optional[str]: ...

    # def process_streaming_message(
    #     self, message: IOStreamingMessage
    # ) -> Optional[str]: ...

    def create_subconversation(self) -> "UI": ...


@runtime_checkable
class WSGI(Protocol):
    def handle_wsgi(
        self,
        app: "Runnable",
        environ: dict[str, Any],
        start_response: Callable[..., Any],
    ) -> list[bytes]: ...


@runtime_checkable
class ASGI(Protocol):
    async def handle_asgi(
        self,
        app: "Runnable",
        scope: dict[str, Any],
        receive: Callable[[dict[str, Any]], Awaitable[None]],
        send: Callable[[dict[str, Any]], Awaitable[None]],
    ) -> None: ...


Workflow = TypeVar("Workflow", bound=Callable[["Workflows", UI, str, str], str])


Agent = TypeVar("Agent")


@runtime_checkable
class Workflows(Protocol):
    def register(
        self, name: str, description: str
    ) -> Callable[[Workflow], Workflow]: ...

    def run(self, name: str, session_id: str, ui: UI, initial_message: str) -> str: ...

    @property
    def names(self) -> list[str]: ...

    def get_description(self, name: str) -> str: ...

    def register_api(
        self,
        api: "OpenAPI",
        callers: Union[Agent, Iterable[Agent]],
        executors: Union[Agent, Iterable[Agent]],
        functions: Optional[
            Union[str, Iterable[Union[str, Mapping[str, Mapping[str, str]]]]]
        ] = None,
    ) -> None: ...


@runtime_checkable
class Runnable(Protocol):
    @contextmanager
    def create(self, import_string: str) -> Generator[None, None, None]: ...

    def start(
        self,
        *,
        import_string: str,
        name: Optional[str] = None,
        initial_message: Optional[str] = None,
        single_run: bool = False,
    ) -> None: ...

    @property
    def wf(self) -> Workflows: ...

    @property
    def ui(self) -> UI: ...

    @property
    def title(self) -> str: ...

    @property
    def description(self) -> str: ...


def run_workflow(
    *,
    wf: Workflows,
    ui: UI,
    name: Optional[str],
    initial_message: Optional[str] = None,
    single_run: bool = False,
) -> None:
    """Run a workflow.

    Args:
        wf (Workflows): The workflows object to use.
        ui (UI): The UI object to use.
        name (Optional[str]): The name of the workflow to run. If not provided, the default workflow will be run.
        initial_message (Optional[str], optional): The initial message to send to the workflow. If not provided, a default message will be sent. Defaults to None.
        single_run (bool, optional): If True, the workflow will only be run once. Defaults to False.
    """
    while True:
        name = wf.names[0] if name is None else name
        description = wf.get_description(name)

        if initial_message is None:
            initial_message = ui.process_message(
                TextInput(
                    sender="FastAgency",
                    recipient="user",
                    prompt=(
                        f"Starting a new workflow '{name}' with the following description:"
                        + "\n\n"
                        + f"{description}"
                        + "\n\nPlease enter an initial message"
                    ),
                )
            )
        else:
            ui.process_message(
                SystemMessage(
                    sender="FastAgency",
                    recipient="user",
                    message={
                        "body": (
                            f"Starting a new workflow '{name}' with the following description:"
                            + "\n\n"
                            + textwrap.indent(description, prefix=" " * 2)
                            + "\n\nand using the following initial message:"
                            + textwrap.indent(initial_message, prefix=" " * 2)
                        )
                    },
                )
            )

        result = wf.run(
            name=name,
            session_id="session_id",
            ui=ui.create_subconversation(),
            initial_message="Hi!" if initial_message is None else initial_message,
        )

        ui.process_message(
            WorkflowCompleted(
                sender="workflow",
                recipient="user",
                result=result,
            )
        )

        initial_message = None

        if single_run:
            break



from pathlib import Path

from fastapi_code_generator.parser import OpenAPIParser
from fastapi_code_generator.visitor import Visitor

from fastagency.api.openapi.security import BaseSecurity


def custom_visitor(parser: OpenAPIParser, model_path: Path) -> dict[str, object]:
    if "securitySchemes" not in parser.raw_obj["components"]:
        return {}
    security_schemes = parser.raw_obj["components"]["securitySchemes"]

    # for k, v in security_schemes.items():
    #     security_schemes[k]["in_value"] = security_schemes[k].pop("in")

    security_classes = []
    security_parameters = {}
    for k, v in security_schemes.items():
        if "in" not in v:
            in_value = v["scheme"] if "scheme" in v and v["type"] == "http" else None
        else:
            in_value = v["in"]
        security_class = BaseSecurity.get_security_class(
            type=v["type"], in_value=in_value
        )
        if security_class is None:
            continue
        security_classes.append(security_class)
        name = v.get("name", None)
        security_parameters[k] = f'{security_class}(name="{name}")'

    return {
        "security_schemes": security_schemes,
        "security_classes": security_classes,
        "security_parameters": security_parameters,
    }


visit: Visitor = custom_visitor



import os
from typing import Any


def openai_llm_config(model: str) -> dict[str, Any]:
    zeros = "0" * 20
    api_key = os.getenv("OPENAI_API_KEY", default=f"sk-{zeros}T3BlbkFJ{zeros}")

    config_list = [
        {
            "model": model,
            "api_key": api_key,
        }
    ]

    llm_config = {
        "config_list": config_list,
        "temperature": 0.0,
    }

    return llm_config


openai_gpt4o_llm_config = openai_llm_config("gpt-4o")

openai_gpt4o_mini_llm_config = openai_llm_config("gpt-4o")



import os

from autogen.agentchat import ConversableAgent, UserProxyAgent

from fastagency import UI, FastAgency
from fastagency.api.openapi.client import OpenAPI
from fastagency.api.openapi.security import APIKeyHeader
from fastagency.runtime.autogen.base import AutoGenWorkflows
from fastagency.ui.console import ConsoleUI

llm_config = {
    "config_list": [
        {
            "model": "gpt-4o-mini",
            "api_key": os.getenv("OPENAI_API_KEY"),
        }
    ],
    "temperature": 0.0,
}


wf = AutoGenWorkflows()


@wf.register(name="weatherman_workflow", description="Weatherman chat")
def simple_workflow(
    wf: AutoGenWorkflows, ui: UI, initial_message: str, session_id: str
) -> str:
    user_proxy = UserProxyAgent(
        name="User_Proxy",
        human_input_mode="ALWAYS",
        llm_config=llm_config,
    )
    weatherman = ConversableAgent(
        name="Weatherman",
        system_message="You are a weatherman.",
        llm_config=llm_config,
    )

    weather_client = OpenAPI.create(
        openapi_url="https://weather.tools.fastagency.ai/openapi.json"
    )
    # Set global security params for all methods
    weather_client.set_security_params(
        APIKeyHeader.Parameters(value="secure weather key")
    )

    wf.register_api(
        api=weather_client,
        callers=user_proxy,
        executors=weatherman,
    )

    chat_result = user_proxy.initiate_chat(
        weatherman,
        message=initial_message,
        summary_method="reflection_with_llm",
        max_turns=3,
    )

    return chat_result.summary


app = FastAgency(wf=wf, ui=ConsoleUI())
